﻿{$include SPIs.inc}

unit SPIBmp;

interface

uses
  Windows, SysUtils, Classes, Graphics, SPIs;

type
  TInSPIAbleData = class(TObject)
  private
    FBMInfoHandle: THandle;
    FBMBitsHandle: THandle;
    FBMInfo: Pointer;
    FBMBits: Pointer;
    FFileName: String;
    FPictureInfo: TPictureInfo;
    FOwner: TObject;
    FOnProgress: TProgressEvent;
    FHasPictureInfo: Boolean;

    //  InSPI に Buf のデーターを読めるかどうかを問い合わせて結果を返す
    function  IsSupported(InSPI: TInSPI; const Buf): Boolean;

    //  読み込みを試行するプラグインの一覧を返す
    function  CreatePluginList: TList;

    //  InSPI 経由で Offset バイト目から読む
    procedure LoadWith(InSPI: TInSPI; const HeadBuf; Offset: Integer);

    //
    function GetBMBits: Pointer;

    //
    function GetBMInfo: Pointer;
  protected
    procedure Progress(Stage: TProgressStage; PercentDone: Byte; const Msg: String);

    //  データの最初の Count バイトを Buf に入れる
    procedure FillHeadBuf(var Buf; Count: Integer); virtual; abstract;

    //  データの Offset バイト目から画像データを読み込む
    procedure LoadPictureInfo(InSPI: TInSPI; Offset: Integer; var Info: TPictureInfo; var Done: Boolean); virtual; abstract;

    //  データの Offset バイト目からPreview で画像を読み込む
    //  読み込めなかったら BMInfo, BMBits は 0 にする
    procedure LoadPreview(InSPI: TInSPI; Offset: Integer; var BMInfo: THandle; var BMBits: THandle); virtual; abstract;

    //  データの Offset バイト目から画像を読み込む
    //  読み込めなかったら BMInfo, BMBits は 0 にする
    procedure LoadPicture(InSPI: TInSPI; Offset: Integer; var BMInfo: THandle; var BMBits: THandle); virtual; abstract;
  public
    constructor Create(Owner: TObject);
    destructor Destroy; override;

    procedure Load;

    procedure Clear;
    function  IsEmpty: Boolean;
    function  PictureInfoIsEmpty: Boolean;

    property  BMInfo: Pointer read GetBMInfo;
    property  BMBits: Pointer read GetBMBits;

    property  FileName: String read FFileName;
    property  PictureInfo: TPictureInfo read FPictureInfo;
    property  OnProgress: TProgressEvent read FOnProgress write FOnProgress;
  end;

  TInSPIAbleFromFile = class(TInSPIAbleData)
  private
  protected
    //  データの最初の Count バイトを Buf に入れる
    procedure FillHeadBuf(var Buf; Count: Integer); override;

    //  データの Offset 目から画像データを読み込む
    procedure LoadPictureInfo(InSPI: TInSPI; Offset: Integer; var Info: TPictureInfo; var Done: Boolean); override;

    //  データの Offset 目からPreview で画像を読み込む
    //  読み込めなかったら BMInfo, BMBits は 0 にする
    procedure LoadPreview(InSPI: TInSPI; Offset: Integer; var BMInfo: THandle; var BMBits: THandle); override;

    //  データの Offset 目から画像を読み込む
    //  読み込めなかったら BMInfo, BMBits は 0 にする
    procedure LoadPicture(InSPI: TInSPI; Offset: Integer; var BMInfo: THandle; var BMBits: THandle); override;
  public
    constructor Create(Owner: TObject; FileName: String);
  end;

  TInSPIAbleFromStream = class(TInSPIAbleData)
  private
    FSource: Pointer;
    FSize: Integer;
    FMemStream: TMemoryStream;
  protected
    //  データの最初の Count バイトを Buf に入れる
    procedure FillHeadBuf(var Buf; Count: Integer); override;

    //  データの Offset 目から画像データを読み込む
    procedure LoadPictureInfo(InSPI: TInSPI; Offset: Integer; var Info: TPictureInfo; var Done: Boolean); override;

    //  データの Offset 目からPreview で画像を読み込む
    //  読み込めなかったら BMInfo, BMBits は 0 にする
    procedure LoadPreview(InSPI: TInSPI; Offset: Integer; var BMInfo: THandle; var BMBits: THandle); override;

    //  データの Offset 目から画像を読み込む
    //  読み込めなかったら BMInfo, BMBits は 0 にする
    procedure LoadPicture(InSPI: TInSPI; Offset: Integer; var BMInfo: THandle; var BMBits: THandle); override;
  public
    constructor Create(Owner: TObject; Stream: TStream; FileName: String);
    destructor Destroy; override;
  end;

function  ProgressCallback(Num, Done:integer; Data: LONG_PTR): Integer; stdcall;

type
  //  Susie プラグインからの読込クラス
  //  内部使用のための最小限の実装しかしていないので，
  //  TGraphic から継承した他のクラスのような汎用性はない
  TSPIGraphic = class(TGraphic)
  private
    FData: TInSPIAbleData;
    FHeight: Integer;
    FWidth: Integer;
    FPalette: HPALETTE;
    FInfoSize: Integer;
    FBitsSize: Integer;
    FColorCount: Integer;
    FPixelFormatIs4bit: Boolean;
    FDecodedDataBMInfo: Boolean;

    class procedure GetOrderedTwoGraphicClasses(var Primaly, Secondly: TGraphicClass; AlternativeClass: TGraphicClass; AlternativeFirst: Boolean);
  protected
    procedure Draw(ACanvas: TCanvas; const Rect: TRect); override;
    function GetEmpty: Boolean; override;
    function GetHeight: Integer; override;
    function GetWidth: Integer; override;
    procedure SetHeight(Value: Integer); override;
    procedure SetWidth(Value: Integer); override;
    procedure NeedDecodeDataBMInfo;
    procedure FreeData;
    procedure Load;
  public
    constructor Create; override;
    destructor Destroy; override;

    function  GetPalette: HPALETTE; {$ifdef VER100OVER} override; {$endif}
    function  PixelFormatIs4bit: Boolean;
    procedure LoadFromStreamWithName(Stream: TStream; const FileName: String);
      virtual;
    procedure LoadFromStream(Stream: TStream); override;
    procedure LoadFromFile(const FileName: String); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure LoadFromClipboardFormat(AFormat: Word; AData: THandle;
                                      APalette: HPALETTE); override;
    procedure SaveToClipboardFormat(var AFormat: Word; var AData: THandle;
                                    var APalette: HPALETTE); override;

    function  CreateAndAssignToDIBFileReadableGraphic(DestClass: TGraphicClass): TGraphic;
    function  CreateAndConvertToDIBFileReadableGraphic(DestClass: TGraphicClass): TGraphic;

    class function CreateFromFile(FileName: String;
      AlternativeClass: TGraphicClass; AlternativeFirst: Boolean;
      ProgressEvent: TProgressEvent): TGraphic;
    class function CreateFromStreamWithName(Stream: TStream; FileName: String;
      AlternativeClass: TGraphicClass; AlternativeFirst: Boolean;
      ProgressEvent: TProgressEvent): TGraphic;
  end;

  TSPIBitmap = class(TBitmap)
  private
    FOnLoadFromStreamWithName: TSPILoadFromStreamWithNameEvent;

    procedure SetOnLoadFromStreamWithName(
      Value: TSPILoadFromStreamWithNameEvent);
{$ifdef VER100OVER}
    procedure SPIGraphicProgress(Sender: TObject; Stage: TProgressStage;
      PercentDone: Byte; RedrawNow: Boolean; const R: TRect; const Msg: String);
{$endif}
  protected
    procedure ConvertFrom(Source: TGraphic); virtual;
  public
    constructor Create; override;

    procedure LoadFromFile(const Filename: String); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure LoadFromStreamWithName(Stream: TStream; const FileName: String);
      virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property  OnLoadFromStreamWithName: TSPILoadFromStreamWithNameEvent
      read FOnLoadFromStreamWithName write SetOnLoadFromStreamWithName;
  end;

const
  SInvalidSPIGraphic    = 'Susie Plugins could not decode';
  SLoadingGraphicBySPI  = 'Loading Graphic by Susie Plugin';
  SSeekingSPI           = 'Seeking enabled Susie Plugins';
  SLoadingBySPI         = 'Loading by [%s]';
  SLoadedGraphicBySPI   = 'Loaded Graphic by Susie Plugin';
  SLoadedGraphic        = 'Loaded Graphic';

implementation

function  StrNPas(Str:PChar; N: Integer): String;
var
  idx: Integer;
begin
  for idx := 0 to N-1 do begin
    if Str[idx] = #0 then begin
      Result := Str;
      Exit;
    end;
  end;
  SetLength(Result, N);
  Move(Str[0], Result[1], N);
end;

procedure LocalFreeAndClear(var Handle: THandle);
begin
  if Handle <> 0 then begin
    LocalFree(Handle);
    Handle := 0;
  end;
end;

procedure LocalUnlockAndPointerNil(Handle: THandle; var P: Pointer);
begin
  if Assigned(P) then begin
    LocalUnlock(Handle);
    P := nil;
  end;
end;

procedure LocalLockIfPointerIsNil(Handle: THandle; var P: Pointer);
begin
  if not Assigned(P) then begin
    P := LocalLock(Handle);
  end;
end;

{ TInSPIAbleData }

procedure TInSPIAbleData.Clear;
begin
  LocalUnlockAndPointerNil(FBMInfoHandle, FBMInfo);
  LocalUnlockAndPointerNil(FBMBitsHandle, FBMBits);

  LocalFreeAndClear(FBMInfoHandle);
  LocalFreeAndClear(FBMBitsHandle);
  LocalFreeAndClear(FPictureInfo.HInfo);
end;

constructor TInSPIAbleData.Create(Owner: TObject);
begin
  inherited Create;

  FOwner := Owner;
end;

function TInSPIAbleData.CreatePluginList: TList;
begin
  Result := TList.Create;
  try
    if SusiePlugins <> nil then begin
      if SusiePlugins.GraphicForcedLoader = nil then begin
        SusiePlugins.FindPluginList([TInSPI], ExtractFileExt(FFileName), Result);
      end
      else begin
        Result.Add(SusiePlugins.GraphicForcedLoader);
      end;
    end;
  except
    Result.Free;
    raise;
  end;
end;

destructor TInSPIAbleData.Destroy;
begin
  Clear;

  inherited;
end;

function TInSPIAbleData.GetBMBits: Pointer;
begin
  LocalLockIfPointerIsNil(FBMBitsHandle, FBMBits);
  Result := FBMBits;
end;

function TInSPIAbleData.GetBMInfo: Pointer;
begin
  LocalLockIfPointerIsNil(FBMInfoHandle, FBMInfo);
  Result := FBMInfo;
end;

function TInSPIAbleData.IsEmpty: Boolean;
begin
  Result := (FBMInfoHandle = 0) or (FBMBitsHandle = 0);
end;

function TInSPIAbleData.IsSupported(InSPI: TInSPI; const Buf): Boolean;
var
  AnsiName : AnsiString;
begin
  if InSPI.EnabledUnicodeAPI then
    Result := InSPI.IsSupportedW(PWideChar(FFileName), @Buf) <> 0
  else begin
    AnsiName:=AnsiString(FFileName);
    Result := InSPI.IsSupported(PAnsiChar(AnsiName), @Buf) <> 0;
  end;
end;

procedure TInSPIAbleData.Load;
const
  headBufSize = 2048+MacBinaryOffset;
var
  headBuf: array [0..headBufSize-1] of byte;
  idx: Integer;
  pluginList: TList;
  spi: TInSPI;
  macBinary: Boolean;
  readMacBinary: Boolean;
begin
  macBinary := (SusiePlugins <> nil) and SusiePlugins.MacBinary;

{$ifdef VER100OVER}
  Progress(psStarting, 0, SLoadingGraphicBySPI);
  try
{$endif}
    Clear;

{$ifdef VER100OVER}
    Progress(psRunning, 1, SSeekingSPI);
{$endif}
    pluginList := CreatePluginList;
    try
      FillChar(headBuf[0], SizeOf(headBuf), 0);
      FillHeadBuf(headBuf, SizeOf(headBuf));
      for idx := 0 to pluginList.Count-1 do begin
        spi := TObject(pluginList[idx]) as TInSPI;
        for readMacBinary := False to macBinary do begin
{$ifdef VER100OVER}
          Progress(psRunning, 1, Format(SLoadingBySPI, [ExtractFileName(spi.FileName)]));
{$endif}
          LoadWith(spi, headBuf, MacBinaryOffsets[readMacBinary]);
          if not IsEmpty then begin
            Exit;
          end;
          Clear;
        end;
      end;
    finally
      pluginList.Free;
    end;
{$ifdef VER100OVER}
  finally
    Progress(psEnding, 100, SLoadedGraphicBySPI);
  end;
{$endif}
end;

procedure TInSPIAbleData.LoadWith(InSPI: TInSPI; const HeadBuf;
  Offset: Integer);
begin
  if not IsSupported(InSPI, PAnsiChar(@HeadBuf)[Offset]) then begin
    Exit;
  end;

  LoadPictureInfo(InSPI, Offset, FPictureInfo, FHasPictureInfo);

  if (SusiePlugins <> nil) and SusiePlugins.GraphicPreview then begin
    LoadPreview(InSPI, Offset, FBMInfoHandle, FBMBitsHandle);
    if IsEmpty then begin
      LoadPicture(InSPI, Offset, FBMInfoHandle, FBMBitsHandle);
    end;
  end
  else begin
    LoadPicture(InSPI, Offset, FBMInfoHandle, FBMBitsHandle);
  end;
end;

function TInSPIAbleData.PictureInfoIsEmpty: Boolean;
begin
  Result := not FHasPictureInfo;
end;

procedure TInSPIAbleData.Progress(Stage: TProgressStage; PercentDone: Byte;
  const Msg: String);
begin
  if Assigned(FOnProgress) then begin
    FOnProgress(FOwner, Stage, PercentDone, Stage = psEnding, Rect(0,0,0,0), Msg);
  end;
end;

{ TInSPIAbleFromFile }

constructor TInSPIAbleFromFile.Create(Owner: TObject; FileName: String);
begin
  inherited Create(Owner);

  FFileName := FileName;
end;

procedure TInSPIAbleFromFile.FillHeadBuf(var Buf; Count: Integer);
var
  stream: TStream;
  streamSize: Integer;
begin
  try
    stream := TFileStream.Create(FFileName, fmOpenRead);
    try
      streamSize := stream.Size;
      if streamSize < Count then begin
        Count := streamSize;
      end;
      stream.Read(Buf, Count);
    finally
      stream.Free;
    end;
  except
    //
  end;
end;

procedure TInSPIAbleFromFile.LoadPicture(InSPI: TInSPI; Offset: Integer;
  var BMInfo, BMBits: THandle);
var
  AnsiName : AnsiString;
begin
  if InSPI.EnabledUnicodeAPI then begin
    if InSPI.GetPictureW(
         PWideChar(FFileName), Offset, 0,
         BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
      BMInfo := 0;
      BMBits := 0;
    end;
  end else begin
    AnsiName:=AnsiString(FFileName);
    if InSPI.GetPicture(
         PAnsiChar(AnsiName), Offset, 0,
         BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
      BMInfo := 0;
      BMBits := 0;
    end;
  end;
end;

procedure TInSPIAbleFromFile.LoadPictureInfo(InSPI: TInSPI; Offset: Integer;
  var Info: TPictureInfo; var Done: Boolean);
var
  AnsiName : AnsiString;
begin
  if InSPI.EnabledUnicodeAPI then
    Done := InSPI.GetPictureInfoW(PWideChar(FFileName), Offset, 0, Info) = 0
  else begin
    AnsiName:=AnsiString(Filename);
    Done := InSPI.GetPictureInfo(PAnsiChar(AnsiName), Offset, 0, Info) = 0;
  end;
end;

procedure TInSPIAbleFromFile.LoadPreview(InSPI: TInSPI; Offset: Integer;
  var BMInfo, BMBits: THandle);
var
  AnsiName : AnsiString;
begin
  try
    if InSPI.EnabledUnicodeAPI then begin
      if InSPI.GetPreviewW(
           PWideChar(FFileName), Offset, 0,
           BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
        BMInfo := 0;
        BMBits := 0;
      end;
    end else begin
      AnsiName:=AnsiString(FFileName);
      if InSPI.GetPreview(
           PAnsiChar(AnsiName), Offset, 0,
           BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
        BMInfo := 0;
        BMBits := 0;
      end;
    end;
  except
    on ESPIInvalidAPI do begin
      BMInfo := 0;
      BMBits := 0;
    end;
  end;
end;

{ TInSPIAbleFromStream }

constructor TInSPIAbleFromStream.Create(Owner: TObject; Stream: TStream; FileName: String);
begin
  inherited Create(Owner);

  if Stream is TMemoryStream then begin
    FSource := TMemoryStream(Stream).Memory;
  end
  else begin
    FMemStream := TMemoryStream.Create;
    FMemStream.LoadFromStream(Stream);
    FSource := FMemStream.Memory;
  end;
  FSize := Stream.Size;
  FFileName := FileName;
end;

destructor TInSPIAbleFromStream.Destroy;
begin
  FMemStream.Free;

  inherited;
end;

procedure TInSPIAbleFromStream.FillHeadBuf(var Buf; Count: Integer);
begin
  if FSize < Count then begin
    Count := FSize;
  end;
  Move(FSource^, Buf, Count);
end;


procedure TInSPIAbleFromStream.LoadPicture(InSPI: TInSPI; Offset: Integer;
  var BMInfo, BMBits: THandle);
begin
  if InSPI.EnabledUnicodeAPI then begin
    if InSPI.GetPictureW(
         PWideChar(FSource), FSize-Offset, 1,
         BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
      BMInfo := 0;
      BMBits := 0;
    end;
  end else begin
    if InSPI.GetPicture(
         PAnsiChar(FSource), FSize-Offset, 1,
         BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
      BMInfo := 0;
      BMBits := 0;
    end;
  end;
end;

procedure TInSPIAbleFromStream.LoadPictureInfo(InSPI: TInSPI; Offset: Integer;
  var Info: TPictureInfo; var Done: Boolean);
begin
  if InSPI.EnabledUnicodeAPI then
    Done := InSPI.GetPictureInfoW(PWideChar(FSource), FSize-Offset, 1, Info) = 0
  else
    Done := InSPI.GetPictureInfo(PAnsiChar(FSource), FSize-Offset, 1, Info) = 0;
end;

procedure TInSPIAbleFromStream.LoadPreview(InSPI: TInSPI; Offset: Integer;
  var BMInfo, BMBits: THandle);
begin
  try
    if InSPI.EnabledUnicodeAPI then begin
      if InSPI.GetPreviewW(
           PWideChar(FSource), FSize-Offset, 1,
           BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
        BMInfo := 0;
        BMBits := 0;
      end;
    end else begin
      if InSPI.GetPreview(
           PAnsiChar(FSource), FSize-Offset, 1,
           BMInfo, BMBits, @ProgressCallback, LONG_PTR(Self)) <> 0 then begin
        BMInfo := 0;
        BMBits := 0;
      end;
    end;
  except
    on ESPIInvalidAPI do begin
      BMInfo := 0;
      BMBits := 0;
    end;
  end;
end;

function  ProgressCallback(Num, Done:Integer; Data: LONG_PTR): Integer;
begin
  Result:=0;
{$ifdef VER100OVER}
  try
    if Data <> 0 then begin
      with TInSPIAbleData(Data) do begin
        if Assigned(FOnProgress) then begin
          Progress(psRunning, MulDiv(Num, 97, Done)+2, '');
        end;
      end;
    end;
  except
    on EAbort do begin
      Result := 1;
    end;
    else begin
      raise;
    end;
  end;
{$endif}
end;


{ TSPIGraphic }

constructor TSPIGraphic.Create;
begin
  inherited;
end;

function TSPIGraphic.CreateAndAssignToDIBFileReadableGraphic(
  DestClass: TGraphicClass): TGraphic;
var
  stream: TMemoryStream;
begin
  Result := DestClass.Create;
  try
    stream := TMemoryStream.Create;
    try
      SaveToStream(stream);
      stream.Position := 0;
      Result.LoadFromStream(stream);
    finally
      stream.Free;
    end;
  except
    Result.Free;
    raise;
  end;
end;

function TSPIGraphic.CreateAndConvertToDIBFileReadableGraphic(
  DestClass: TGraphicClass): TGraphic;
var
  stream: TMemoryStream;
begin
{$ifndef Ver140Over}
  Result := TGraphic(TBareGraphicClass(DestClass).Create);
{$else}
  Result := DestClass.Create;
{$endif}
  try
    stream := TMemoryStream.Create;
    try
      SaveToStream(stream);
      FreeData;
      stream.Position := 0;
      Result.LoadFromStream(stream);
    finally
      stream.Free;
    end;
  except
    Result.Free;
    raise;
  end;
end;

class function TSPIGraphic.CreateFromFile(FileName: String;
  AlternativeClass: TGraphicClass; AlternativeFirst: Boolean;
  ProgressEvent: TProgressEvent): TGraphic;
var
  classes: array [0..1] of TGraphicClass;
  idx: Integer;
begin
  GetOrderedTwoGraphicClasses(classes[0], classes[1], AlternativeClass, AlternativeFirst);

  for idx := 0 to 1 do begin
    Result := classes[idx].Create;
    try
      Result.OnProgress := ProgressEvent;
      Result.LoadFromFile(FileName);
      Break;
    except
      Result.Free;
      Result := nil;
    end;
  end;
end;

class function TSPIGraphic.CreateFromStreamWithName(Stream: TStream;
  FileName: String; AlternativeClass: TGraphicClass;
  AlternativeFirst: Boolean;
  ProgressEvent: TProgressEvent): TGraphic;
var
  classes: array [0..1] of TGraphicClass;
  idx: Integer;
  curPos: Integer;
begin
  curPos := Stream.Position;

  GetOrderedTwoGraphicClasses(classes[0], classes[1], AlternativeClass, AlternativeFirst);

  for idx := 0 to 1 do begin
    Result := classes[idx].Create;
    try
      Result.OnProgress := ProgressEvent;
      if classes[idx] = Self then begin
        TSPIGraphic(Result).LoadFromStreamWithName(Stream, FileName);
      end
      else begin
        Result.LoadFromStream(Stream);
      end;
      Break;
    except
      Result.Free;
      Result := nil;
    end;
    Stream.Position := curPos;
  end;
end;

destructor TSPIGraphic.Destroy;
begin
  FreeData;

  inherited;
end;

procedure TSPIGraphic.Draw(ACanvas: TCanvas; const Rect: TRect);
var
  oldPalette: HPALETTE;
  newPalette: HPALETTE;
  oldStretchMode: Integer;
  paletteSelected: Boolean;
begin
  if Assigned(FData) then begin
    NeedDecodeDataBMInfo;

    paletteSelected := False;
    oldPalette := 0;
    newPalette := Palette;
    if newPalette <> 0 then begin
      oldPalette := SelectPalette(ACanvas.Handle, newPalette, True);
      RealizePalette(ACanvas.Handle);
      paletteSelected := True;
    end;

    oldStretchMode := SetStretchBltMode(ACanvas.Handle, COLORONCOLOR);

    StretchDIBits(ACanvas.Handle,
      Rect.Left, Rect.Top, Rect.Right-Rect.Left, Rect.Bottom-Rect.Top,
      0, 0, FWidth, FHeight,
      FData.BMBits, PBitmapInfo(FData.BMInfo)^,
      DIB_RGB_COLORS, ACanvas.CopyMode);

    SetStretchBltMode(ACanvas.Handle, oldStretchMode);

    if paletteSelected then begin
      SelectPalette(ACanvas.Handle, oldPalette, True);
    end;
  end;
end;

procedure TSPIGraphic.FreeData;
begin
  FData.Free;
  FData := nil;

  if FPalette <> 0 then begin
    DeleteObject(FPalette);
    FPalette := 0;
  end;

  FDecodedDataBMInfo := False;
end;

function TSPIGraphic.GetEmpty: Boolean;
begin
  Result := not Assigned(FData) or FData.IsEmpty;
end;

function TSPIGraphic.GetHeight: Integer;
begin
  NeedDecodeDataBMInfo;

  Result := FHeight;
end;

class procedure TSPIGraphic.GetOrderedTwoGraphicClasses(var Primaly,
  Secondly: TGraphicClass; AlternativeClass: TGraphicClass; AlternativeFirst: Boolean);
begin
  if AlternativeFirst then begin
    Primaly  := AlternativeClass;
    Secondly := Self;
  end
  else begin
    Primaly  := Self;
    Secondly := AlternativeClass;
  end;
end;

function TSPIGraphic.GetPalette: HPALETTE;
begin
  NeedDecodeDataBMInfo;

  Result := FPalette;
end;

function TSPIGraphic.GetWidth: Integer;
begin
  NeedDecodeDataBMInfo;

  Result := FWidth;
end;

procedure TSPIGraphic.Load;
begin
  try
    FData.OnProgress   := Self.OnProgress;
    FData.Load;
    if FData.IsEmpty then begin
      raise EInvalidGraphic.Create(SInvalidSPIGraphic);
    end;
  except
    FreeData;
    raise;
  end;
  Changed(Self);
end;

procedure TSPIGraphic.LoadFromClipboardFormat(AFormat: Word;
  AData: THandle; APalette: HPALETTE);
begin
  //
end;

procedure TSPIGraphic.LoadFromFile(const FileName: String);
begin
  FreeData;
  FData := TInSPIAbleFromFile.Create(Self, FileName);
  Load;
end;

procedure TSPIGraphic.LoadFromStream(Stream: TStream);
begin
  LoadFromStreamWithName(Stream, '');
end;

procedure TSPIGraphic.LoadFromStreamWithName(Stream: TStream;
  const FileName: String);
begin
  FreeData;
  FData := TInSPIAbleFromStream.Create(Self, Stream, FileName);
  Load;
end;

procedure TSPIGraphic.NeedDecodeDataBMInfo;
var
  pBMInfo: PBitmapInfo;
  bitCount: Integer;
  logPalette: PLogPalette;
  colorIdx: Integer;
begin
  if FDecodedDataBMInfo then begin
    Exit;
  end;

  If Empty then begin
    FWidth := 0;
    FHeight := 0;
    FColorCount := 0;
    FPalette := 0;
    FBitsSize := 0;
    FInfoSize := 0;
    FPixelFormatIs4bit := False;
    FDecodedDataBMInfo := True;
    Exit;
  end;

  pBMInfo := PBitmapInfo(FData.BMInfo);
  with pBMInfo^.bmiHeader do begin
    bitCount    := biBitCount;
    FColorCount := biClrUsed;
    FBitsSize   := ((bitCount*biWidth+31) div 32)*4*abs(biHeight);
    FWidth      := biWidth;
    FHeight     := biHeight;
  end;

  if FColorCount = 0 then begin
    case bitCount of
      1: FColorCount := 2;
      4: FColorCount := 16;
      8: FColorCount := 256;
    end;
  end;

  FInfoSize := SizeOf(TBitmapInfoheader)+FColorCount*SizeOf(TRGBQuad);
  FPixelFormatIs4bit := bitCount = 4;

  GetMem(logPalette, SizeOf(TLogPalette)+SizeOf(TPaletteEntry)*FColorCount);
  try
    with logPalette^, pBMInfo^  do begin
      palVersion := $300;
      palNumEntries := FColorCount;
      for colorIdx := 0 to FColorCount-1 do begin
        palPalEntry[colorIdx].peRed 	:= bmiColors[colorIdx].rgbRed;
        palPalEntry[colorIdx].peGreen := bmiColors[colorIdx].rgbGreen;
        palPalEntry[colorIdx].peBlue 	:= bmiColors[colorIdx].rgbBlue;
        palPalEntry[colorIdx].peFlags := 0;
      end;
    end;
    FPalette := CreatePalette(logPalette^);
  finally
    FreeMem(logPalette);
  end;
  FDecodedDataBMInfo := True;
end;

function TSPIGraphic.PixelFormatIs4bit: Boolean;
begin
  NeedDecodeDataBMInfo;

  Result := FPixelFormatIs4bit;
end;

procedure TSPIGraphic.SaveToClipboardFormat(var AFormat: Word;
  var AData: THandle; var APalette: HPALETTE);
begin
  //
end;

procedure TSPIGraphic.SaveToStream(Stream: TStream);
var
  bmFH: TBitmapFileheader;
begin
  NeedDecodeDataBMInfo;

  FillChar(bmFH, SizeOf(bmFH), 0);
  with bmFH do begin
    bfType    := $4D42;
    bfSize    := SizeOf(bmFH)+FInfoSize+FBitsSize;
    bfOffBits := SizeOf(bmFH)+FInfoSize;
  end;

  Stream.WriteBuffer(bmFH, sizeOf(bmFH));
  Stream.WriteBuffer(FData.BMInfo^, FInfoSize);
  Stream.WriteBuffer(FData.BMBits^, FBitsSize);
end;

procedure TSPIGraphic.SetHeight(Value: Integer);
begin
  //
end;

procedure TSPIGraphic.SetWidth(Value: Integer);
begin
  //
end;

{------TSPIBitmap------}
procedure TSPIBitmap.Assign(Source: TPersistent);
var
  newGraphic: TGraphic;
begin
  if Source is TSPIGraphic then begin
    newGraphic := TSPIGraphic(Source).CreateAndAssignToDIBFileReadableGraphic(TGraphicClass(ClassParent));
    try
      inherited Assign(newGraphic);
    finally
      newGraphic.Free;
    end;
{$ifdef VER100OVER}
    //  4bit画像時に勝手に TBitmap がシステムパレットに変えやがる・・・
    //  ではなくて、変えていただいている件への対策
    //  なんで、わざわざ、パレットを作って代入し直さなきゃならんのだ・・・(--;;
    //  何を考えてるのか、よくわからんぞ＞ VCL
    //	D3 以降のみ有効っす。
    //	D2 じゃこの方法はダメみたい。
    if TSPIGraphic(Source).PixelFormatIs4bit then begin
      Palette := CopyPalette(TSPIGraphic(Source).Palette);
    end;
{$endif}
    Exit;
  end;
  inherited Assign(Source);
end;

procedure TSPIBitmap.ConvertFrom(Source: TGraphic);
var
  tempPalette: HPALETTE;
var
  newGraphic: TGraphic;
begin
  try
    if not Assigned(Source) then begin
      raise EInvalidGraphic.Create(SInvalidSPIGraphic);
    end
    else if Source is TSPIGraphic then begin
      if TSPIGraphic(Source).PixelFormatIs4bit then begin
        tempPalette := CopyPalette(TSPIGraphic(Source).Palette);
      end
      else begin
        tempPalette := 0;
      end;
      newGraphic := TSPIGraphic(Source).CreateAndConvertToDIBFileReadableGraphic(TGraphicClass(Self.ClassParent));
      try
        Assign(newGraphic);
      finally
        newGraphic.Free;
      end;
      if tempPalette <> 0  then begin
        Palette := tempPalette;
      end;
    end
    else begin
      Assign(Source);
    end;
  finally
    Source.Free;
  end;
end;

constructor TSPIBitmap.Create;
begin
  inherited Create;

  FOnLoadFromStreamWithName := LoadFromStreamWithName;
end;

procedure TSPIBitmap.LoadFromFile(const Filename: String);
begin
  try
    ConvertFrom(
      TSPIGraphic.CreateFromFile(
        FileName,
        TGraphicClass(Self.ClassParent),
        AnsiUpperCase(ExtractFileExt(FileName))='.BMP',
        SPIGraphicProgress
      )
    );
  finally
    Progress(Self, psEnding, 100, True, Rect(0, 0, Width, Height), SLoadedGraphic);
  end;
end;


procedure TSPIBitmap.LoadFromStream(Stream: TStream);
begin
  LoadFromStreamWithName(Stream, '');
end;

procedure TSPIBitmap.LoadFromStreamWithName(Stream: TStream; const
  FileName: String);
begin
  try
    ConvertFrom(
      TSPIGraphic.CreateFromStreamWithName(
        Stream,
        FileName,
        TGraphicClass(Self.ClassParent),
        AnsiUpperCase(ExtractFileExt(FileName)) = '.BMP',
        SPIGraphicProgress
      )
    );
  finally
    Progress(Self, psEnding, 100, True, Rect(0, 0, Width, Height), SLoadedGraphic);
  end;
end;

procedure TSPIBitmap.SetOnLoadFromStreamWithName(
  Value: TSPILoadFromStreamWithNameEvent);
begin
  //
end;

procedure TSPIBitmap.SPIGraphicProgress(Sender: TObject;
  Stage: TProgressStage; PercentDone: Byte; RedrawNow: Boolean;
  const R: TRect; const Msg: String);
begin
  if Stage <> psEnding then begin
    Progress(Self, Stage, PercentDone, RedrawNow, R, Msg);
  end;
end;

initialization
  TSPIs.SetDefaultGraphic(TSPIBitmap);
finalization
  TPicture.UnregisterGraphicClass(TSPIBitmap);
end.

