{$include SPIs.inc}

unit SPINkDIB;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, SPIs, SPIBmp, NkDIB;

type
  TSPINkDIBPaletteOption = (sdpNone, sdpDev8bit, sdpAllways);

  TSPINkDIB = class(TNkDIB)
  private
    FPaletteOption: TSPINkDIBPaletteOption;
    FOnLoadFromStreamWithName: TSPILoadFromStreamWithNameEvent;

    procedure SetOnLoadFromStreamWithName(
      Value: TSPILoadFromStreamWithNameEvent);
    procedure SPIGraphicProgress(Sender: TObject; Stage: TProgressStage;
      PercentDone: Byte; RedrawNow: Boolean; const R: TRect; const Msg: string);
  protected
    procedure Changed(Sender: TObject); override;
    procedure ConvertFrom(Source: TGraphic); virtual;
  public
    constructor Create; override;

    procedure LoadFromFile(const Filename: String); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure LoadFromStreamWithName(Stream: TStream; const FileName: String);
      virtual;

    procedure Assign(Source: TPersistent); override;
    property  PaletteOption: TSPINkDIBPaletteOption
      read FPaletteOption write FPaletteOption;
  published
    property  OnLoadFromStreamWithName: TSPILoadFromStreamWithNameEvent
      read FOnLoadFromStreamWithName write SetOnLoadFromStreamWithName;
  end;


implementation

var
  Dev8bit: Boolean;

{------TSPINkDIB------}

procedure TSPINkDIB.Assign(Source: TPersistent);
var
  newGraphic: TGraphic;
begin
  if Source is TSPIGraphic then begin
    newGraphic := TSPIGraphic(Source).CreateAndAssignToDIBFileReadableGraphic(TGraphicClass(ClassParent));
    try
      inherited Assign(newGraphic);
    finally
      newGraphic.Free;
    end;
    Exit;
  end
  else if Source is TSPINkDIB then begin
    FPaletteOption := TSPINkDIB(Source).FPaletteOption;
  end;
  inherited Assign(Source);
end;

procedure TSPINkDIB.Changed(Sender: TObject);
var
  makePalette: Boolean;
begin
  //  NkDIB  16/32bpp ǍނƁA24bpp ɕϊ̂ŁA
  //  BitCount  16,32 ɂ͂Ȃ̂ǁA̔łł
  //  16,32 ɂȂ邩Ȃ̂
  if (BitCount in [16, 24, 32]) and
     (Palette = 0) then begin
    makePalette := False;
    case FPaletteOption of
      sdpAllways: begin
        makePalette := True;
      end;
      sdpDev8bit: begin
        makePalette := Dev8bit;
      end;
    end;
    if makePalette then begin
      CreateTrueColorPalette;
    end;
  end;
  inherited Changed(Sender);
end;

procedure TSPINkDIB.ConvertFrom(Source: TGraphic);
var
  newGraphic: TGraphic;
begin
  try
    if not Assigned(Source) then begin
      raise EInvalidGraphic.Create(SInvalidSPIGraphic);
    end
    else if Source is TSPIGraphic then begin
      newGraphic := TSPIGraphic(Source).CreateAndConvertToDIBFileReadableGraphic(TGraphicClass(Self.ClassParent));
      try
        Assign(newGraphic);
      finally
        newGraphic.Free;
      end;
    end
    else begin
      Assign(Source);
    end;
  finally
    Source.Free;
  end;
end;

constructor TSPINkDIB.Create;
begin
  inherited Create;

  FPaletteOption := sdpDev8bit;

  FOnLoadFromStreamWithName := LoadFromStreamWithName;
end;

procedure TSPINkDIB.LoadFromFile(const Filename: String);
begin
  try
    ConvertFrom(
      TSPIGraphic.CreateFromFile(
        FileName,
        TGraphicClass(Self.ClassParent),
        AnsiUpperCase(ExtractFileExt(FileName))='.BMP',
        SPIGraphicProgress
      )
    );
  finally
    Progress(Self, psEnding, 100, True, Rect(0, 0, Width, Height), SLoadedGraphic);
  end;
end;


procedure TSPINkDIB.LoadFromStream(Stream: TStream);
begin
  LoadFromStreamWithName(Stream, '');
end;


procedure TSPINkDIB.LoadFromStreamWithName(Stream: TStream; const
  FileName: String);
begin
  try
    ConvertFrom(
      TSPIGraphic.CreateFromStreamWithName(
        Stream,
        FileName,
        TGraphicClass(Self.ClassParent),
        UpperCase(ExtractFileExt(FileName)) = '.BMP'
        {$ifdef VER100OVER}, SPIGraphicProgress {$endif}
      )
    );
  finally
    Progress(Self, psEnding, 100, True, Rect(0, 0, Width, Height), SLoadedGraphic);
  end;
end;

procedure TSPINkDIB.SetOnLoadFromStreamWithName(
  Value: TSPILoadFromStreamWithNameEvent);
begin
//
end;


procedure TSPINkDIB.SPIGraphicProgress(Sender: TObject;
  Stage: TProgressStage; PercentDone: Byte; RedrawNow: Boolean;
  const R: TRect; const Msg: String);
begin
  if Stage <> psEnding then begin
    Progress(Self, Stage, PercentDone, RedrawNow, R, Msg);
  end;
end;

var
  dc: THandle;


initialization
  Dev8bit := False;
  dc := GetDC(0);
  if dc <> 0 then begin
    try
      Dev8bit := (GetDeviceCaps(dc, Planes)*GetDeviceCaps(dc,BitsPixel) <= 8);
    finally
      ReleaseDC(0, dc);
    end;
  end;
finalization
  TPicture.UnregisterGraphicClass(TSPINkDIB);
end.
